searchData={"items":[{"type":"module","title":"Ack","doc":"`Ack` is a tiny application that might be used as drop-in to handle\n  acknowledgments between processes.\n\nWhen `App1` sends _something_ to `App2` it might require the acknowledgement\n  back confirming the successful processing of this _something_. For instance,\n  `App1` might call an API endpoint of `App2`, which triggers a long process,\n  or it might place a message into RabbitMQ and expect to receive an `ack` to\n  perform some cleanup, or whatever.\n\nIn this scenario, `App1` might instruct `Ack` to listen for the `ack` from\n  `App2` for a message\n\n- `%{key: key, timeout: msecs}`\n\nwhere `key` is the unique identifier of the message to be acknowledged.\nUpon receival, if the `key` is known to the system, `Ack` will broadcast\nthe message of the following shape\n\n- `%{status: :ack, key: key}`\n\nto `:ack` channel. `App1` should be subscribed to `{Ack.Horn, :ack}` channel\nin order to receive this message.\n\nIf the `key` is unknown to the system, on of the following possible messages\n\n- `%{status: :nack, key: key}`, when the `key` was explicitly not acked\n- `%{status: :invalid, key: key}`, when the `key` is not known to the system\n- `%{status: :unknown, key: key, value: value}`, when somewhat unexpected happened\n\nThe former one os routed to `{Ack.Horn, :nack}` channel, the last two\nare broadcasted to `{Ack.Horn, :error}` channel.\n\nThe broadcast is done with [`Envío`](https://hexdocs.pm/envio/envio.html)\n  package, consider reading the documentation there to get more details about\n  subscriptions.\n\nCurrently only HTTP endpoint is supported for ack callbacks.","ref":"Ack.html"},{"type":"function","title":"Ack.listen/1","doc":"Adds a listener for the key with a timeout specified (defaults to `5` sec.)","ref":"Ack.html#listen/1"},{"type":"type","title":"Ack.t/0","doc":"","ref":"Ack.html#t:t/0"},{"type":"module","title":"Ack.Horn","doc":"The publisher, sending broadcasts when `ACK` is received. It serves three\n  different channels:\n\n- `{Ack.Horn, :ack}` to send broadcast when the `ACK` comes\n- `{Ack.Horn, :nack}` to send broadcast when the `NACK` comes\n- `{Ack.Horn, :error}` to send broadcast when the client:\n  - called back with the wrong key (`status: :invalid`)\n  - called back with the unknown `ACK` value (`status: :unknown`)\n  - did not called back and the timeout has expired.\n\nHost and client might agree on using more verbs besides `ACK` and `NACK`.\nTo handle this the host should implement somewhat like the following clause:\n\n```elixir\ndef handle_envio(%{status: :unknown, key: key, value: verb} = message, state) do\n  state = BusinessLogic.on_other_verb(key)\n  {:noreply, state}\nend\n```","ref":"Ack.Horn.html"},{"type":"function","title":"Ack.Horn.broadcast/1","doc":"","ref":"Ack.Horn.html#broadcast/1"},{"type":"extras","title":"Ack With Ease","doc":"# ![Ack Logo](https://raw.githubusercontent.com/am-kantox/ack/master/stuff/i/logo-48.png)   Ack With Ease","ref":"how_to_ack.html"},{"type":"extras","title":"What? - Ack With Ease","doc":"`Ack` is designed to be a lightweight easy-to-use implementation of _interprocess acknowledgement_, whatever it means. According to [_Wikipedia_](https://en.wikipedia.org/wiki/Acknowledgement_(data_networks)):\n\n> In data networking, telecommunications, and computer buses, an acknowledgement (_ACK_) is a signal passed between communicating processes, computers, or devices to signify acknowledgement, or receipt of message, as part of a communications protocol. The negative-acknowledgement (_NAK_ or _NACK_) signal is sent to reject a previously received message, or to indicate some kind of error. Acknowledgements and negative acknowledgements inform a sender of the receiver's state so that it can adjust its own state accordingly.\n\nThat said, when the independently running process is to pass a message around to another process—via some complicated pipe—that makes it hard to ensure that all parts of this pipe delivered the message successfully and/or instead of propagating the _success_/_failure_ back through each link in this chain, the sender might simply require _ACK_ back when the message was successfully delivered to the recipient.\n\nThis could involve the additional level of complexity when we are talking about tiny microservices. Not each and every microservice are to have a bucket of intercommunication clients included. Lightweight ones might have none in fact. This package is a drop-in solution for such cases.\n\n_It probably does not apply when all involved processes come with messaging queue and fully-functional web server on board._","ref":"how_to_ack.html#what"},{"type":"extras","title":"Why? - Ack With Ease","doc":"This package comes with all the boilerplate included, including lightweight web-server based on [`Camarero`](https://hexdocs.pm/camarero/camarero.html) and process message broadcasting with [`Envío`](https://hexdocs.pm/envio). It drastically simplifies adding _ACK_ support to lightweight microservices. Indeed, all one needs would be to:\n\n- add `Ack` to the list of applications started with `App1`\n- implement [`Envio.Subscriber`](https://hexdocs.pm/envio/Envio.Subscriber.html) in `App1`, listening to any of three following channels:\n  - `{Ack.Horn, :ack}`\n  - `{Ack.Horn, :nack}`\n  - `{Ack.Horn, :error}`\n- implement `App2` to `HTTP POST` to `App1.domain:30009` one of two requests (assuming `key` is somewhat negotiated upgront and known to `App1`):\n  - `%{\"key\" => key, \"value\" => \"ack\"}` to `ack`, or\n  - `%{\"key\" => key, \"value\" => \"nack\"}` to `nack`\n\nThe control flow of each _ACK_ would be as shown below.\n\n![Ack Message Lifetime](https://raw.githubusercontent.com/am-kantox/ack/master/stuff/i/ack.png)\n\nThe _whole_ implementation required from `App1` to start supporting _ACKs_ would look like (besides the business logic handling _ACKs_ and _NACKs_):\n\n```elixir\ndefmodule MyApp.AckHandler do\n  use Envio.Subscriber,\n    channels: [{Ack.Horn, :ack}, {Ack.Horn, :nack}, {Ack.Horn, :error}]\n\n  def handle_envio(%{key: key, status: status} = message, state) do\n    state = BusinessLogic.on_response(status, for: key)\n    {:noreply, state}\n  end\nend\n```\n\nTo start expecting for the _ACK_, one simple calls `Ack.listen/1` passing the following map as a parameter of a type `Ack.t`:\n\n```elixir\nAck.listen(%{key: \"user_123\", timeout: 5_000})\n```\n\nWhen the callback with a payload having `%{}` will be received, the message will be broadcasted to all the pub-sub listeners of one of aforementioned `Ack.Horn` channels, depending on the state.\n\n---\n\nThe _whole_ implementation required from `App2` to start supporting _ACKs_ would be to _HTTP POST_ to the predefined endpoint the message of a shape `%{key: entity_id, value: :ack or :nack}`.\n\nThat’s it.","ref":"how_to_ack.html#why"},{"type":"extras","title":"Why not? - Ack With Ease","doc":"In many [over]complicated systems there are already robust queue-based ack-back pipelines presented. If the whole system already has the intercommunication pipeline, providing back-and-forth validation and monitoring of everything, `Ack` would most likely not suit. Still, it’s robust and comprehensive.","ref":"how_to_ack.html#why-not"},{"type":"extras","title":"Is it of any good? - Ack With Ease","doc":"Sure, it is.","ref":"how_to_ack.html#is-it-of-any-good"},{"type":"extras","title":"Plugging Ack In","doc":"# ![Ack Logo](https://raw.githubusercontent.com/am-kantox/ack/master/stuff/i/logo-48.png)   Plugging Ack In","ref":"plugging_it_in.html"},{"type":"extras","title":"Standalone Cowboy - Plugging Ack In","doc":"When the _ack-back_ functionality is to be added to the standalone microservice, having no webserver on board, `Ack` (in fact, underlying [`Camarero`](https://hexdocs.pm/camarero/camarero.html) instance) should be configured to start `cowboy` server underneath. Put this in your mix configuration:\n\n```elixir\nconfig :camarero,\n  cowboy: [port: 4001, options: []], # options are passed to cowboy plug\n  carta: [Ack.Callback, Ack.Active], # list of handlers/endpoints\n  root: \"api/v2\"                     # root location to serve requests\n```","ref":"plugging_it_in.html#standalone-cowboy"},{"type":"extras","title":"Keyword options - Plugging Ack In","doc":"#### `cowboy:`\n\nFor the list of options that can be passed, see [`plug_cowboy` docs](https://hexdocs.pm/plug_cowboy/Plug.Cowboy.html#module-options).\n\n#### `carta:`\n\nHere we might specify any amount of different [`camarero` handlers](https://hexdocs.pm/camarero/camarero.html#handlers). `[Ack.Callback, Ack.Active]` is the default value for those, but one might easily everwrite/extend this list with other endpoints.\n\n- `Ack.Callback` is the handler/endpoint for callbacks, accepting `POST` requests of a shape `%{key:  , value:   |  }` only, available at the path `\"#{root}/callback\"`.\n- `Ack.Active` accepts `GET` only and exports the map of callbacks currently being awaited, available at the path `\"#{root}/active\"`.\n\n#### `root:`\n\nThe root location where the plug handler(s) will be served.","ref":"plugging_it_in.html#keyword-options"},{"type":"extras","title":"Listener - Plugging Ack In","doc":"`Ack` comes with it’s own [`Registry`](https://hexdocs.pm/elixir/Registry.html), provided by [`Envio`](https://hexdocs.pm/envio/).\n\nTo place the callback into the map of active expectations, call `Ack.listen/1`. It accepts a map, having two parameters, mandatory `key` and optional `timeout`, defaulted to `5_000` (ms.)\n\n---\n\nWhen the `POST` request is issued to the callback endpoint, it’s being processed as shown below:\n\n- if the `key` is not known to `Ack` _or_ if the `value` is neither `\"ack\"` nor `\"nack\"`, the broadcast is sent to all the subscribers (see below) through `{Ack.Horn, :error}` channel, with all the corresponding data.\n- if the `value` is `\"ack\"`, the respective key _is removed_ from the _expecting callbacks_ map and the broadcast is sent through `{Ack.Horn, :ack}` channel.\n- if the `value` is `\"nack\"`, the respective key _is **not** removed_ from the _expecting callbacks_ map and the broadcast is sent through `{Ack.Horn, :nack}` channel.\n\nIf there was no callback within the desired timeout, the broadcast with `status: :timeout` is being sent through `{Ack.Horn, :error}` and the respective key _is **not** removed_ from the _expecting callbacks_ map (this might be changed in the future.)\n\n#### List of possible statuses\n\n- `ack` (`{Ack.Horn, :ack}` channel) the remote issued `ack`, _everything is fine_\n- `nack` (`{Ack.Horn, :nack}`) the remote issued `nack`, meaning it expects us to resend the data\n- `invalid` (`{Ack.Horn, :error}`) there is no such `key` in the list of active expectaions\n- `unknown` (`{Ack.Horn, :error}`) the remote sent the unknown term (nether `nack` nor `ack`)\n- `timeout` (`{Ack.Horn, :error}`) the remote did not respond withing the specified timeout","ref":"plugging_it_in.html#listener"},{"type":"extras","title":"Subscriber(s) - Plugging Ack In","doc":"To subscribe to the callback notifications, one should implement an [`Envio.Subscriber`](https://hexdocs.pm/envio/Envio.Subscriber.html) behaviour. The easiest way would be to use scaffold:\n\n```elixir\ndefmodule MyApp.AckHandler do\n  use Envio.Subscriber,\n    channels: [{Ack.Horn, :ack}, {Ack.Horn, :nack}, {Ack.Horn, :error}]\n\n  def handle_envio(%{key: key, status: status} = message, state) do\n    state = BusinessLogic.on_response(status, for: key)\n    {:noreply, state}\n  end\nend\n```","ref":"plugging_it_in.html#subscriber-s"},{"type":"extras","title":"The summing up - Plugging Ack In","doc":"The application hosting `Ack` implements one or more subscribers as shown above. When the `ACK` is required, it places it to the list of expectations with a call to `Ack.listen/1`. When the `ACK` arrives, or after the timeout (what comes earlier) all the subscribers do receive _an envío_.\n\nThe client application performs an `HTTP` post to the respective endpoint with a payload of a shape `%{key:  , value:   |  }`. The contracting parties might use _other_ types of `ACK` statuses, assuming the host handles the respective `handle_envio(%{status: status}, _)` properly.","ref":"plugging_it_in.html#the-summing-up"},{"type":"extras","title":"Phoenix Integration - Plugging Ack In","doc":"To use `Ack` with [`Phoenix`](https://phoenixframework.org/), one should follow the steps above. The only difference would be, instead of spawning `cowboy`, we should add this line to _Phoenix_ `routes.ex` file:\n\n```elixir\nforward(\"/api/ack\", Ack.Camarero.Handler)\n```\n\nThe `root` option of _Camarero_ config is now relative to the first parameter in the call above. To keep routing simple, we suggest the setting `confir :camarero, root: \"\"`.\n\nNow `/api/ack/callback` would be the _callback_ endpoint and `/api/ack/active` would be the _active_ endpoint respectively.\n\nThat’s it!","ref":"plugging_it_in.html#phoenix-integration"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,51,46,48]}}